from __future__ import annotations

import asyncio
import os
from typing import TYPE_CHECKING
from unittest.mock import Mock
from uuid import uuid4

import pytest
import pytest_asyncio

from cashews import Cache
from cashews.backends.memory import Memory
from cashews.backends.transaction import TransactionBackend

if TYPE_CHECKING:  # pragma: no cover
    from cashews.backends.interface import Backend


pytest_plugins = ["pytest_asyncio"]  # pylint: disable=invalid-name

try:
    import aiohttp
except ImportError:
    pass
else:
    pytest_plugins.append("aiohttp.pytest_plugin")

    del aiohttp


@pytest.fixture(scope="session")
def event_loop():
    """Create an instance of the default event loop for session."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def redis_dsn():
    host = os.getenv("REDIS_HOST", "")
    port = os.getenv("REDIS_PORT", "6379")
    return f"redis://{host}:{port}/"


@pytest.fixture(scope="session")
def backend_factory():
    def factory(backend_cls: type[Backend], *args, **kwargs):
        backend = backend_cls(*args, **kwargs)
        return backend

    return factory


@pytest_asyncio.fixture(
    name="raw_backend",
    scope="session",
    params=[
        "memory",
        "transactional",
        pytest.param("redis", marks=pytest.mark.redis),
        pytest.param("redis_hash", marks=pytest.mark.redis),
        pytest.param("redis_cs", marks=pytest.mark.redis),
        pytest.param("diskcache", marks=pytest.mark.diskcache),
    ],
)
async def _backend(request, redis_dsn, backend_factory):
    if request.param == "diskcache":
        from cashews.backends.diskcache import DiskCache

        backend = backend_factory(DiskCache, shards=0)
    elif request.param == "redis":
        from cashews.backends.redis import Redis

        backend = backend_factory(Redis, redis_dsn, hash_key=None, max_connections=20, safe=False, socket_timeout=10)
    elif request.param == "redis_hash":
        from cashews.backends.redis import Redis

        backend = backend_factory(
            Redis, redis_dsn, hash_key=uuid4().hex, max_connections=20, safe=False, socket_timeout=10
        )
    elif request.param == "redis_cs":
        from cashews.backends.redis.client_side import BcastClientSide

        backend = backend_factory(
            BcastClientSide, redis_dsn, hash_key=None, max_connections=5, safe=False, socket_timeout=10
        )
        backend._expire_for_recently_update = 0.1
    elif request.param == "transactional":
        backend = TransactionBackend(backend_factory(Memory))
    else:
        backend = backend_factory(Memory, check_interval=0.01)
    try:
        await backend.init()
        yield backend
    finally:
        await backend.close()


@pytest_asyncio.fixture()
async def backend(raw_backend):
    await raw_backend.clear()
    yield raw_backend


@pytest.fixture(scope="session")
def target(raw_backend):
    return Mock(wraps=raw_backend, is_full_disable=False, name=str(raw_backend))


@pytest_asyncio.fixture()
async def cache(target: Mock):
    await target.clear()
    target.reset_mock()
    cache = Cache()
    cache._add_backend(target)
    return cache



import inspect
from functools import wraps
from typing import Optional, Union

from cashews._typing import TTL, AsyncCallable_T, Decorator, KeyOrTemplate
from cashews.backends.interface import _BackendInterface
from cashews.exceptions import LockedError
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

__all__ = ("locked",)


def locked(
    backend: _BackendInterface,
    key: Optional[KeyOrTemplate] = None,
    ttl: Optional[TTL] = None,
    max_lock_ttl: int = 10,
    min_wait_time: Optional[TTL] = None,
    step: Union[float, int] = 0.1,
    prefix: str = "lock",
) -> Decorator:
    """
    Decorator that can help you to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    Lock following function calls till first one will be finished
    Can guarantee that one function call for given ttl, if ttl is None

    :param backend: cache backend
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param ttl: duration to lock wrapped function call
    :param max_lock_ttl: default ttl if it not set
    :param min_wait_time: minimum time to wait till lock is released (if ttl not set)
    :param step: duration between lock check
    :param prefix: custom prefix for key, default 'lock'
    """
    ttl = ttl_to_seconds(ttl)

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        _min_wait_time = ttl_to_seconds(min_wait_time)
        if inspect.isasyncgenfunction(func):
            return _asyncgen_lock(func, backend, ttl, _key_template, max_lock_ttl, _min_wait_time, step)
        return _coroutine_lock(func, backend, ttl, _key_template, max_lock_ttl, _min_wait_time, step)

    return _decor


def _coroutine_lock(
    func: AsyncCallable_T,
    backend: _BackendInterface,
    ttl: Optional[TTL],
    key_template: KeyOrTemplate,
    max_lock_ttl: int,
    min_wait_time: Optional[int],
    step: Union[float, int],
) -> AsyncCallable_T:
    @wraps(func)
    async def _wrap(*args, **kwargs):
        _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
        _cache_key = get_cache_key(func, key_template, args, kwargs)
        try:
            async with backend.lock(_cache_key, _ttl or max_lock_ttl):
                return await func(*args, **kwargs)
        except LockedError:
            if not await backend.is_locked(_cache_key, wait=_ttl or min_wait_time, step=step):
                return await func(*args, **kwargs)
            raise

    return _wrap


def _asyncgen_lock(
    func: AsyncCallable_T,
    backend: _BackendInterface,
    ttl: Optional[TTL],
    key_template: KeyOrTemplate,
    max_lock_ttl: int,
    min_wait_time: Optional[int],
    step: Union[float, int],
) -> AsyncCallable_T:
    @wraps(func)
    async def _wrap(*args, **kwargs):
        _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
        _cache_key = get_cache_key(func, key_template, args, kwargs)
        try:
            async with backend.lock(_cache_key, _ttl or max_lock_ttl):
                async for chunk in func(*args, **kwargs):
                    yield chunk
                return
        except LockedError:
            if not await backend.is_locked(_cache_key, wait=_ttl or min_wait_time, step=step):
                async for chunk in func(*args, **kwargs):
                    yield chunk
                return
            raise

    return _wrap



import random

import pytest

from cashews import Cache

pytestmark = [pytest.mark.integration]


@pytest.fixture(
    name="app",
    scope="session",
    params=[
        "memory",
        pytest.param("redis", marks=pytest.mark.integration),
        pytest.param("redis_cs", marks=pytest.mark.integration),
        pytest.param("diskcache", marks=pytest.mark.integration),
    ],
)
def _app(request, redis_dsn):
    dsn = "mem://"
    if request.param == "diskcache":
        dsn = "disk://"
    elif request.param == "redis":
        dsn = redis_dsn
    elif request.param == "redis_cs":
        dsn = redis_dsn + "&client_side=t"
    cache = Cache()
    cache.setup(dsn)

    from fastapi import FastAPI, Header
    from starlette.responses import StreamingResponse

    app = FastAPI()

    @app.get("/")
    @cache.early(ttl="1s", key="root:{q}")
    async def root(q: str = "q", x_name: str = Header("test")):
        return {"name": x_name, "q": q}

    @cache.iterator(ttl="5s")
    async def iterator():
        for i in range(10):
            yield f"{random.randint(0, 9)}"

    @app.get("/stream")
    async def stream():
        return StreamingResponse(iterator(), status_code=201, headers={"X-Test": "TRUE"})

    return app, cache


@pytest.fixture(name="client", scope="session")
def _client(app):
    from starlette.testclient import TestClient

    with TestClient(app[0]) as client:
        yield client


def test_no_cache(client):
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"name": "test", "q": "q"}

    response = client.get("/?q=test", headers={"X-Name": "name"})
    assert response.status_code == 200
    assert response.json() == {"name": "name", "q": "test"}


def test_cache(client):
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"name": "test", "q": "q"}

    response = client.get("/", headers={"X-Name": "name"})
    assert response.status_code == 200
    assert response.json() == {"name": "test", "q": "q"}


def test_cache_stream(client, app):
    response = client.get("/stream")
    assert response.status_code == 201
    assert response.headers["X-Test"] == "TRUE"
    was = response.content

    response = client.get("/stream")
    assert response.status_code == 201
    assert response.headers["X-Test"] == "TRUE"
    assert response.content == was
